/* Zone file database manipulation.
   This file is part of WebDNS.
 */

/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */

/* Since the customer has access to the name server (through WebDNS) to
   manipulate zone files, we need to keep a fast lookup database to
   try and prevent the customer adding a domain that already exists. This
   will happen if the customer tries to add the same zone file to two
   virtual servers or if the customer is trying to add a domain that is
   owned by another account. 
 */

#include "config.h"
#include <dirent.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#ifdef HAVE_GDBM_H
#include <gdbm.h>
#elif defined(HAVE_NDBM_H)
#include <ndbm.h>
#endif

#include "libvsd.h"
#include "vsd.h"
#include "bind.h"


/* Open the bind database.  Return a database handle.  */
void *bind_dbase_open (int op, const char *db)
{
  struct stat sb;
  void *status = NULL;

#if 0
  if (stat (db, &sb))
    return status;
#endif

#ifdef HAVE_GDBM_H
  if (op == DB_WRITE)
    status = gdbm_open ((char *) db, 0, GDBM_WRITER, 0600, NULL);
  else if (op == DB_READ)
    status = gdbm_open ((char *) db, 0, GDBM_READER, 0400, NULL);
  else if (op == DB_WRCREAT)
    status = gdbm_open ((char *) db, 0, GDBM_NEWDB, 0600, NULL);
#elif defined(HAVE_NDBM_H)
  if (op == DB_WRITE)
    status = dbm_open ((char *) db, O_EXCL | O_RDWR, 0600);
  else if (op == DB_READ)
    status = dbm_open ((char *) db, DBM_RDONLY, 0400);
  else if (op == DB_WRCREAT)
    status = dbm_open ((char *) db, O_CREAT | O_TRUNC | O_RDWR | O_EXCL, 0600);
#endif
  return status;
}

/* Close the bind database.  */
void bind_dbase_close (void *dbf)
{
#ifdef HAVE_GDBM_H
  gdbm_close ((GDBM_FILE) dbf);
#elif defined(HAVE_NDBM_H)
  dbm_close ((DBM *) dbf);
#endif
}

/* Lookup `domain' in the database.  Return 1 if it exists and 0 if it does
   not.  */
int bind_dbase_lookup (void *dbf, const char *domain)
{
  datum key;

  key.dptr = (char *) domain;
  key.dsize = strlen (domain) + 1;

#ifdef HAVE_GDBM_H
  return gdbm_exists ((GDBM_FILE) dbf, key) ? 1 : 0;
#elif defined(HAVE_NDBM_H)
  {
    datum result = dbm_fetch ((DBM *) dbf, key);
    return (result.dptr == NULL) ? 0 : 1;
  }
#endif
}

/* Delete `domain' from the database.  */
void bind_dbase_delete (void *dbf, const char *domain)
{
  datum key;

  key.dptr = (char *) domain;
  key.dsize = strlen (domain) + 1;

#ifdef HAVE_GDBM_H
  gdbm_delete ((GDBM_FILE) dbf, key);
#elif defined(HAVE_NDBM_H)
  dbm_delete ((DBM *) dbf, key);
#endif
}

/* Delete `domain' from the database.  */
void bind_dbase_add (void *dbf, const char *domain)
{
  datum key, content;

  key.dptr = (char *) domain;
  key.dsize = strlen (domain) + 1;

  content.dptr = "";
  content.dsize = 0;
#ifdef HAVE_GDBM_H
  gdbm_store ((GDBM_FILE) dbf, key, content, GDBM_INSERT);
#elif defined(HAVE_NDBM_H)
  dbm_store ((DBM *) dbf, key, content, DBM_INSERT);
#endif
}

static datum bind_dbase_firstkey (void *dbf)
{
#ifdef HAVE_GDBM_H
  return gdbm_firstkey ((GDBM_FILE) dbf);
#elif defined(HAVE_NDBM_H)
  return dbm_firstkey ((DBM *) dbf);
#endif
}

static datum bind_dbase_nextkey (void *dbf, datum key)
{
#ifdef HAVE_GDBM_H
  return gdbm_nextkey ((GDBM_FILE) dbf, key);
#elif defined(HAVE_NDBM_H)
  key = key;
  return dbm_nextkey ((DBM *) dbf);
#endif
}

static char *bind_dbase_strerror (void *dbf)
{
#ifdef HAVE_GDBM_H
  dbf = dbf;
  return gdbm_strerror (gdbm_errno);
#elif defined(HAVE_NDBM_H)
  if (dbf)
    return strerror (dbm_error ((DBM *) dbf));
  else
    return "unknown error";
#endif
}

/* Scan /var/named, adding all zone files to the lookup database.
   We assume that all zone files begin `db.'.  */
void bind_dbase_generate (struct connection *vc)
{
  struct stat sb;
  char **dv;
  int dc;

  if (stat ("/var/named", &sb))
    {
      io_buffer_store (vc, "-ERR /var/named does not exist");
      return;
    }

  if (S_ISDIR (sb.st_mode))
    {
      void *dbf = bind_dbase_open (DB_WRCREAT, BIND_DB);
      if (dbf)
	{
	  int x;
	  for (x = 0; x < vc->map->servers; x++)
	    {
	      vc->server_root = vsd_map_server_root (vc->map,
						     vc->map->server[x].name);
	      dv = bind_dir_get_domains (vc, &dc);

	      while (--dc >= 0)
		{
		  /* Insert domain name into database.  */
		  bind_dbase_add (dbf, dv[dc]);
		  free (dv[dc]);
		}
	      free (dv);
	    }
	  bind_dbase_close (dbf);
	}
      else
	io_buffer_store (vc, "-ERR Cannot create %s: %s", BIND_DB,
			 bind_dbase_strerror (dbf));
    }
  else
    io_buffer_store (vc, "-ERR /var/named is not a directory");
}

void bind_dbase_excl_list (struct connection *vc, int argc, char *argv[])
{
  void *dbf;
  struct stat sb;

  if (stat (BIND_EXCL_DB, &sb))
    {
      io_buffer_store (vc, "-ERR Cannot open %s: db not found",
		       BIND_EXCL_DB);
      return;
    }

  dbf = bind_dbase_open (DB_READ, BIND_EXCL_DB);
  if (! dbf)
    io_buffer_store (vc, "-ERR Cannot open %s: %s", BIND_EXCL_DB,
		     bind_dbase_strerror (dbf));
  else
    {
      datum key, nextkey;

      key = bind_dbase_firstkey (dbf);
      while (key.dptr)
	{
	  io_buffer_store (vc, "%.*s\n", key.dsize, key.dptr);
	  nextkey = bind_dbase_nextkey (dbf, key);
	  free (key.dptr);
	  key = nextkey;
	}

      bind_dbase_close (dbf);
    }
}

void bind_dbase_excl_add (struct connection *vc, int argc, char *argv[])
{
  void *dbf;
  struct stat sb;

  if (stat (BIND_EXCL_DB, &sb))
    {
      int fd = open (BIND_EXCL_DB, O_CREAT, 0755);
      close (fd);
    }

  dbf = bind_dbase_open (DB_WRITE, BIND_EXCL_DB);
  if (! dbf)
    io_buffer_store (vc, "-ERR Cannot open %s: %s", BIND_EXCL_DB,
		     bind_dbase_strerror (dbf));
  else
    {
      bind_dbase_add (dbf, argv[0]);
      bind_dbase_close (dbf);
    }
}

void bind_dbase_excl_delete (struct connection *vc, int argc, char *argv[])
{
  void *dbf;
  struct stat sb;

  if (stat (BIND_EXCL_DB, &sb))
    {
      io_buffer_store (vc, "-ERR Cannot open %s: db not found",
		       BIND_EXCL_DB);
      return;
    }

  dbf = bind_dbase_open (DB_WRITE, BIND_EXCL_DB);
  if (! dbf)
    io_buffer_store (vc, "-ERR Cannot open %s: %s", BIND_EXCL_DB,
		     bind_dbase_strerror (dbf));
  else
    {
      bind_dbase_delete (dbf, argv[0]);
      bind_dbase_close (dbf);
    }
}

/* Check domain exists in database `db'.  Return 1 if not found
   or 0 if found or -1 if the database doesn't exist.  */
int bind_dbase_check (const char *db, const char *domain)
{
  void *dbf = bind_dbase_open (DB_READ, db);

  if (dbf)
    {
      int result = bind_dbase_lookup (dbf, domain);
      bind_dbase_close (dbf);
      return result;
    }

  /* Couldn't open database.  */
  return -1;
}
