/* Zone file manipulation routines.
   This file is part of WebDNS. 
 */

/*
 * VSD - Virtual Server Daemon
 * Copyright (c) 2000 - 2001 Idaya Ltd.
 *
 * VSD is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * VSD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VSD - see the COPYING file;  if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * See the AUTHORS file for a list of contributors to VSD.
 * See the ChangeLog files for a list of changes.
 */


/* #define DEBUG */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#ifdef DEBUG
#include <sys/syslog.h>
#endif

#include "vsd.h"
#include "bind.h"

/* A DNS zone file looks something like this:

   $ORIGIN brain.eu.org.
   @ IN SOA ns0.serve.co.uk. root.ns0.serve.co.uk. (
      1999041900 ; Serial Number YYYYMMDDSS (SS=sequence)
      10800 ; Standard values for other servers caching entries
      3600 ; from this domain
      3600000 ;
      86400 )

                 IN      NS     ns0.serve.co.uk.  ; A name server record
	         IN      NS     ns0.serve.net.uk. ; A name server record
           IN    MX      5 brain.eu.org.
           IN    A       195.89.127.29
    www    IN    A       195.89.127.29
    smtp   IN    CNAME   www
    mail   IN    CNAME   www
*/

extern int yynerrs;

static char current_origin[128];
static char current_domain[128];

void zone_set_origin (const char *new_origin)
{
  /* If the new origin ends with a dot then it is absolute,
     otherwise the old origin is appended to the end.  */
  if (new_origin[strlen (new_origin) - 1] == '.')
    strcpy (current_origin, new_origin);
  else
    {
      char temp[256];

      strcpy (temp, current_origin);
      strcat (current_origin, new_origin);
      strcat (current_origin, ".");
      strcat (current_origin, temp);
    }  
}

void zone_set_domain (const char *new_domain)
{
  /* @ is shorthand notation for the origin.  */
  if (new_domain[0] == '@' && new_domain[1] == '\0')
    strcpy (current_domain, current_origin);
  else
    {
      if (new_domain[strlen (new_domain) - 1] == '.')
	strcpy (current_domain, new_domain);
      else
	{
	  /* Append the current origin.  */
	  strcpy (current_domain, new_domain);
	  strcat (current_domain, ".");
	  strcat (current_domain, current_origin);
	}
    }
}

char *zone_fixup_domain (const char *domain, int set)
{
  char *fixed;

  /* If domain is blank then use last specified domain.  */
  if (domain == NULL)
    return strdup (current_domain);

  if (set)
    {
      /* Domain is not blank so this becomes the new last specified
	 domain.  */
      zone_set_domain (domain);
    }

  /* @ is shorthand for use current origin.  */
  if (domain[0] == '@' && domain[1] == '\0')
    return strdup (current_origin);

  /* If domain ends with a dot then just copy, otherwise
     append the current origin.  */
  if (domain[strlen (domain) - 1] == '.')
    return strdup (domain);

  fixed = (char *) malloc (strlen (domain)
			   + strlen (current_origin) + 2);
  strcpy (fixed, domain);
  strcat (fixed, ".");
  strcat (fixed, current_origin);

  return fixed;
}

/* Add a resource record of type `type' to the zone file `zone_file'.
   If `pos' is >= 0 then replace an old (previously deleted) array position
   into which to add the record otherwise create a new entry.

   `origin', `class', `server' and `...' are the resource record parameters
   and depend on the record `type'.  */
int zone_add_resource_record (struct zonefile *zone_file, int pos,
			      enum resource_type type,
			      const char *domain, const char *class,
			      const char *server, ...)
{
  struct resource_rec rec;
  va_list ap;

  va_start (ap, server);

  rec.type = type;

  rec.domain = zone_fixup_domain (domain, 1);
  rec.class = (class) ? strdup (class) : NULL;

  if (server && ! strcspn (server, "0123456789."))
    rec.server = strdup (server); /* IP address.  */
  else
    rec.server = zone_fixup_domain (server, 0); /* Name.  */

  switch (type)
    {
    case rr_gid:
    case rr_uid:
      rec.extra.gid.gid = va_arg (ap, int);
      break;
    case rr_hinfo:
      rec.extra.hinfo.name1 = strdup (va_arg (ap, char *));
      rec.extra.hinfo.name2 = strdup (va_arg (ap, char *));
      break;
    case rr_mx:
      rec.extra.mx.cost = va_arg (ap, int);
      break;
    case rr_soa:
      rec.extra.soa.primary_master_ns = strdup (va_arg (ap, char *));
      rec.extra.soa.mail_address = strdup (va_arg (ap, char *));
      rec.extra.soa.serial = va_arg (ap, unsigned int);
      rec.extra.soa.refresh = va_arg (ap, int);
      rec.extra.soa.retry = va_arg (ap, int);
      rec.extra.soa.expire = va_arg (ap, int);
      rec.extra.soa.min_ttl = va_arg (ap, int);

      /* If a default TTL for the zone has not been set then use the
	 TTL of the SOA.  */
      if (zone_file->default_ttl == -1)
	zone_file->default_ttl = rec.extra.soa.min_ttl;
      break;
    default:
      break;
    }

  if (pos == -1)
    {
      if (zone_file->rsvc == 0)
	{
	  zone_file->rsv = (struct resource_rec **)
	    malloc (sizeof (struct resource_rec *));
	  zone_file->rsvc = 1;
	}
      else
	{
	  zone_file->rsvc ++;
	  zone_file->rsv = (struct resource_rec **)
	    realloc (zone_file->rsv,
		     zone_file->rsvc * sizeof (struct resource_rec *));
	}
      
      zone_file->rsv[zone_file->rsvc - 1] = (struct resource_rec *)
	malloc (sizeof (struct resource_rec));

      memcpy (zone_file->rsv[zone_file->rsvc - 1], &rec,
	      sizeof (struct resource_rec));
    }
  else
    {
      zone_file->rsv[pos] = (struct resource_rec *)
	malloc (sizeof (struct resource_rec));
      memcpy (zone_file->rsv[pos], &rec, sizeof (struct resource_rec));
    }

  va_end (ap);
  return 0;
}

int zone_find_resource_record (struct zonefile *zone_file,
			       enum resource_type type,
			       const char *domain, const char *class,
			       const char *server, va_list ap)
{
  int x;
  struct resource_rec *rec;
  char *canon_domain, *canon_server;

  /* Canonicalise domains for comparison purposes.  Internally, we
     represent the all resource records in their canonicalised form.
     However, the client might pass us a `domain' or `server' in
     the other form.  We check for both because zone_fixup_domain
     doesn't cope with IP addresses.  */
  canon_domain = zone_fixup_domain (domain, 0);
  canon_server = zone_fixup_domain (server, 0);

#ifdef DEBUG
  syslog (LOG_WARNING, "find: domain = %s, class = %s, server = %s",
	  domain, class, server);
#endif

  for (x = 0; x < zone_file->rsvc; x++)
    {
      /* Skip undefined records.  */
      if (zone_file->rsv[x] == NULL)
	continue;

      /* Look for a resource record of same type.  */
      if (zone_file->rsv[x]->type == type)
	{
	  rec = zone_file->rsv[x];

#ifdef DEBUG
	  syslog (LOG_WARNING, "rec->domain = %s, class = %s, server = %s",
		  rec->domain, rec->class, rec->server);
#endif
	  if (strcmp (domain, rec->domain) != 0
	      && strcmp (canon_domain, rec->domain) != 0)
	    continue;
	  if (strcmp (class, rec->class))
	    continue;
	  if (strcmp (server, rec->server) != 0
	      && strcmp (canon_server, rec->server) != 0)
	    continue;

	  switch (rec->type)
	    {
	    case rr_gid:
	    case rr_uid:
	      {
		int gid = va_arg (ap, int);
		if (rec->extra.gid.gid != gid)
		  continue;
	      }
	      break;
	    case rr_hinfo:
	      {
		/* extra params: <name1> <name2> */
		char *name1 = va_arg (ap, char *);
		char *name2 = va_arg (ap, char *);

		if (strcmp (rec->extra.hinfo.name1, name1)
		    || strcmp (rec->extra.hinfo.name2, name2))
		  continue;
	      }
	      break;
	    case rr_mx:
	      {
		/* extra params: <preference> */
		int cost = va_arg (ap, int);
		if (rec->extra.mx.cost != cost)
		  continue;
	      }
	      break;
	    case rr_soa:
	      {
		/* extra params: <primary master ns> <mail addr>
		   <serial> <refresh> <retry> <expire> <min_ttl>  */
		char *pri_ns = va_arg (ap, char *);
		char *mail = va_arg (ap, char *);
		unsigned int serial = va_arg (ap, unsigned int);
		int refresh = va_arg (ap, int);
		int retry = va_arg (ap, int);
		int expire = va_arg (ap, int);
		int min_ttl = va_arg (ap, int);

		if (strcmp (rec->extra.soa.primary_master_ns, pri_ns)
		    || strcmp (rec->extra.soa.mail_address, mail)
		    || rec->extra.soa.serial != serial
		    || rec->extra.soa.refresh != refresh
		    || rec->extra.soa.retry != retry
		    || rec->extra.soa.expire != expire
		    || rec->extra.soa.min_ttl != min_ttl)
		  continue;
	      }
	      break;
	    default:
	      break;
	    }
	  /* Found a match.  */
	  free (canon_domain);
	  free (canon_server);
	  return x;
	}
    }

  /* No match found.  */
  free (canon_domain);
  free (canon_server);
  return -1;
}

int zone_del_resource_record (struct zonefile *zone_file,
			      enum resource_type type,
			      const char *domain, const char *class,
			      const char *server, ...)
{
  int x;
  va_list ap;

  va_start (ap, server);
  x = zone_find_resource_record (zone_file, type, domain, class, server, ap);
  va_end (ap);

#ifdef DEBUG
  syslog (LOG_WARNING, "zone_find_resource_record: return = %d", x);
#endif

  if (x >= 0)
    {
      zone_free_resource_record (zone_file->rsv[x]);
      zone_file->rsv[x] = NULL;
    }

  return x;
}

/* Read a zone file into memory, parsing it and writing the fields
   into a series of structures.  */
struct zonefile *zone_read (const char *zone_file)
{
  FILE *stream;
  struct zonefile *zone;

  stream = fopen (zone_file, "r");
  if (stream == NULL)
    return NULL;

  zone = (struct zonefile *) malloc (sizeof (struct zonefile));
  memset (zone, 0, sizeof (struct zonefile));
  /* If the file doesn't have a $TTL then we can use the TTL of the
     SOA resource record.  */
  zone->default_ttl = -1;

  yyparse_init (stream, zone);
  /* yydebug = 1; */
  while (yyparse () != 0)
    ;

  fclose (stream);

  if (yynerrs > 0)
    {
      zone_free (zone);
      return NULL;
    }

  yyparse_final ();

  return zone;
}

/* Strip the origin from `domain' if need be.  */
static const char *strip_origin (const char *origin, const char *domain)
{
  static char buf[128];
  char *p;

  if (strcmp (origin, domain) == 0)
    {
      buf[0] = '@';
      buf[1] = '\0';
      return buf;
    }

  p = strstr (domain, origin);
  if (p == NULL)
    return domain;

  if (strcmp (p, origin) == 0 && p[-1] == '.')
    {
      /* `origin' is the last part of the domain name. This is the
	 only case where we will remove it.  */
      strncpy (buf, domain, p - 1 - domain);
      buf[p - 1 - domain] = '\0';
      
      return buf;
    }

  return domain;
}

/* Store a zone file to disk. Add a $ORIGIN and strip `domain_name' from
   each resource record.  */
int zone_save (struct zonefile *zone, const char *zone_file,
	       const char *domain_name)
{
  FILE *stream;
  int x;

  stream = fopen (zone_file, "w");
  if (stream == NULL)
    return -1;

  fprintf (stream, "$TTL %d\n", zone->default_ttl);
  fprintf (stream, "$ORIGIN %s", domain_name);
  if (domain_name[strlen (domain_name) - 1] != '.')
    fputc ('.', stream);
  fputc ('\n', stream);

  for (x = 0; x < zone->rsvc; x++)
    {
      struct resource_rec *rec = zone->rsv[x];

      /* Skip undefined records.  */
      if (rec == NULL)
	continue;

      if (rec->domain)
        fprintf (stream, "%s ", strip_origin (domain_name, rec->domain));

      fputc ('\t', stream);

      if (rec->class)
        fprintf (stream, "%s ", rec->class);

      fputc ('\t', stream);
      switch (rec->type)
	{
 	case rr_a:
	case rr_cname:
 	case rr_mb:
 	case rr_mg:
 	case rr_minfo:
 	case rr_mr:
 	case rr_ns:
 	case rr_ptr:
	  fprintf (stream, "%s\t%s\n",
		   zone_type_to_name (rec->type),
		   strip_origin (domain_name, rec->server));
	  break;
	case rr_gid:
	case rr_uid:
	  fprintf (stream, "%s\t%d\n",
		   zone_type_to_name (rec->type), rec->extra.gid.gid);
	  break;
	case rr_hinfo:
	  fprintf (stream, "HINFO\t%s %s\n", rec->extra.hinfo.name1,
		   rec->extra.hinfo.name2);
	  break;
	  break;
	case rr_mx:
	  fprintf (stream, "MX\t%d %s\n", rec->extra.mx.cost,
		   strip_origin (domain_name, rec->server));
	  break;
	case rr_soa:
	  fprintf (stream, "SOA %s %s (\n",
		   rec->extra.soa.primary_master_ns,
		   rec->extra.soa.mail_address);
	  fprintf (stream, "\t\t%d ; Serial number YYYYMMDDSS (SS=seq)\n",
		   rec->extra.soa.serial);
	  fprintf (stream, "\t\t%d ; Refresh\n",
		   rec->extra.soa.refresh);
	  fprintf (stream, "\t\t%d ; Retry\n",
		   rec->extra.soa.retry);
	  fprintf (stream, "\t\t%d ; Expire\n",
		   rec->extra.soa.expire);
	  fprintf (stream, "\t\t%d ) ; Min TTL\n\n",
		   rec->extra.soa.min_ttl);
	  break;
	default:
	  break;
	}
    }

  fclose (stream);

  return 0;
}

static void freemem (void *buffer)
{
  if (buffer)
    free (buffer);
}

int zone_free_resource_record (struct resource_rec *rec)
{
  if (rec == NULL)
    {
      errno = EINVAL;
      return -1;
    }

  freemem (rec->domain);
  freemem (rec->class);
  freemem (rec->server);
  switch (rec->type)
    {
    case rr_hinfo:
      freemem (rec->extra.hinfo.name1);
      freemem (rec->extra.hinfo.name2);
      break;
    case rr_soa:
      freemem (rec->extra.soa.primary_master_ns);
      freemem (rec->extra.soa.mail_address);
      break;
    default:
      break;
    }
  freemem (rec);

  return 0;
}

void zone_free (struct zonefile *zone_file)
{
  int x;

  for (x = 0; x < zone_file->rsvc; x++)
    {
      /* Skip undefined records.  */
      if (zone_file->rsv != NULL)
	zone_free_resource_record (zone_file->rsv[x]);
    }

  freemem (zone_file->rsv);
  freemem (zone_file);
}

void zone_update_serial (struct zonefile *zone)
{
  int x;

  /* Fix serial numbers on SOA records.  */
  for (x = 0; x < zone->rsvc; x++)
    if (zone->rsv[x] && zone->rsv[x]->type == rr_soa)
      zone->rsv[x]->extra.soa.serial
	= zone_gen_serial (zone->rsv[x]->extra.soa.serial);
}

/* Make a serial number for an SOA record from the current date.  */
unsigned int zone_gen_serial (unsigned int old_serial)
{
  char temp[16];
  unsigned int serial;
  time_t curtime = time (0);

  /* The serial number is in the formay YYYYMMDDSS, where SS is a
     sequence number.  */
  strftime (temp, sizeof (temp), "%Y%m%d", localtime (&curtime));
  serial = atoi (temp);

  if (old_serial != -1)
    {
      /* If we are modifying the zone on the same day as the last
	 modification then just increment the sequence number.  */
      if (serial == (old_serial / 100))
	serial = old_serial + 1;
      else
	serial *= 100;
    }
  else
    serial *= 100;

  return serial;
}

const char *zone_type_to_name (enum resource_type type)
{
  switch (type)
    {
    case rr_a:
      return "A";
    case rr_cname:
      return "CNAME";
    case rr_gid:
      return "GID";
    case rr_hinfo:
      return "HINFO";
    case rr_mb:
      return "MB";
    case rr_mg:
      return "MG";
    case rr_minfo:
      return "MINFO";
    case rr_mr:
      return "MR";
    case rr_mx:
      return "MX";
    case rr_ns:
      return "NS";
    case rr_ptr:
      return "PTR";
    case rr_soa:
      return "SOA";
    case rr_uid:
      return "UID";
    }

  return NULL;
}

enum resource_type zone_name_to_type (const char *name)
{
  if (strcmp (name, "A") == 0)
    return rr_a;
  if (strcmp (name, "CNAME") == 0)
    return rr_cname;
  if (strcmp (name, "GID") == 0)
    return rr_gid;
  if (strcmp (name, "HINFO") == 0)
    return rr_hinfo;
  if (strcmp (name, "MB") == 0)
    return rr_mb;
  if (strcmp (name, "MG") == 0)
    return rr_mg;
  if (strcmp (name, "MINFO") == 0)
    return rr_minfo;
  if (strcmp (name, "MR") == 0)
    return rr_mr;
  if (strcmp (name, "MX") == 0)
    return rr_mx;
  if (strcmp (name, "NS") == 0)
    return rr_ns;
  if (strcmp (name, "PTR") == 0)
    return rr_ptr;
  if (strcmp (name, "SOA") == 0)
    return rr_soa;
  if (strcmp (name, "UID") == 0)
    return rr_uid;

  return -1;
}
