/* Module command interface.
   Copyright (c) 1999, 2000, 2001 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include "module.h"
#include "mod_part.h"
#include "part.h"

static struct command_tag clist[] =
{
  {
    CMD_PART_SKELADD, ADMIN_HOST,
    "Register a new skel within an existing partition",
    "PART_SKELADD", 2, "<skel name> <partition>",
    "part_skeladd", 3, "<vsd server> <skel name> <partition>"
  },
  {
    CMD_PART_SKELDEL, ADMIN_HOST,
    "Unregister an existing skel from a partition",
    "PART_SKELDEL", 2, "<skel name> <partition>",
    "part_skeldel", 3, "<vsd server> <skel name> <partition>"
  },
  {
    CMD_PART_SETSKELGLOBAL, ADMIN_HOST,
    "Register the skel 'skel name' to be the global skel",
    "PART_SETSKELGLOBAL", 1, "<skel name>",
    "part_setskelglobal", 2, "<vsd server> <skelname>"
  },
  {
    CMD_PART_DELSKELGLOBAL, ADMIN_HOST,
    "Unregister the skel 'skel name' from the global section",
    "PART_DELSKELGLOBAL", 1, "<skel name>",
    "part_delskelglobal", 2, "<vsd server> <skel name>"
  },
  { 
    CMD_PART_SKELLIST, ADMIN_HOST | ADMIN_VS,
    "List all of the defined skels",
    "PART_SKELLIST", 0, "[<partition>]",
    "part_skellist", 1, "<vsd server> [<partition>]",
  },
  {
    CMD_PART_SETSKELVS, ADMIN_HOST,
    "Set the skel definition for a virtual server",
    "PART_SETSKELVS", 2, "<virtual server> <skel name>",
    "part_setskelvs", 3, "<vsd server> <virtual server> <skel name>",
  },
  {
    0, 0, NULL, NULL, 0, NULL, NULL, 0, NULL
  }
};

/* Return -1 if `cmd' does not exist.
   Return 1 if there is a syntax error.
   Return 0 if everything is cool.  */
int mod_part_check_cmd (int level, const char *cmd, int argc, char *argv[],
		      struct command_tag **list)
{
  if (cmd == NULL)
    {
      /* Return a pointer to all commands.  */
      *list = clist;
      return 0;
    }

  /* Find command in clist.  */
  *list = module_lookup_command (clist, level, cmd);
  if (*list == NULL)
    return -1;  /* Command does not exist.  */

  /* Check command syntax.  Pretty basic at the moment.  */
#ifdef VSD_LIB
  if (argc < (*list)->vsdadm_args)
    return 1; /* Syntax error.  */
#else
  if (argc < (*list)->vsd_args)
    return 1; /* Syntax error.  */
#endif

  /* Syntax is ok.  */
  return 0;
}

#ifdef VSD_LIB
char *mod_part_cmd (void *io, int level, const char *cmd,
		  int argc, char **argv)
{
  struct command_tag *list;
  char *status = NULL;

  list = module_lookup_command (clist, level, cmd);
  if (! list)
    return "notexist";

  vsd_set_level (io, list->level);
  switch (list->token)
    {
    case CMD_PART_SKELADD:
      vsd_send_command (io, NULL, "PART_SKELADD %s %s",
                        argv[1] /* skel name */, argv[2] /* partition */);
      break;
    case CMD_PART_SKELDEL:
      vsd_send_command (io, NULL, "PART_SKELDEL %s %s",
                        argv[1] /*skel name */, argv[2] /* partition */);
      break;
    case CMD_PART_SETSKELGLOBAL:
      vsd_send_command (io, NULL, "PART_SETSKELGLOBAL %s",
                        argv[1] /* skel name */);
      break;
    case CMD_PART_DELSKELGLOBAL:
      vsd_send_command (io, NULL, "PART_DELSKELGLOBAL %s",
                        argv[1] /* skel name */);
      break;
    case CMD_PART_SKELLIST:
      vsd_send_command (io, NULL, "PART_SKELLIST %s",
                        argv[1]);
      break;
    case CMD_PART_SETSKELVS:
      vsd_send_command (io, NULL, "PART_SETSKELVS %s %s",
                        argv[1] /* vs name */, argv[2] /* skel name */);
      break;
    default:
      return "mod_part: command not recognised";
      break;
    }

  status = vsd_process_reply (io);
  return status;
}

#else /* ! VSD_LIB */
/* Return 0 if command processed, 1 if not recognised.  */
int mod_part_cmd (struct connection *vc, int level, const char *command,
		int argc, char *argv[], const char *data)
{
  int token;
  struct command_tag *tag;

  tag = module_lookup_command (clist, level, command);
  if (tag == NULL)
    return 1;

  token = tag->token;
  switch (token)
    {
    case CMD_PART_SKELADD:
      part_register_skel (vc, argc, argv);
      break;
    case CMD_PART_SKELDEL:
      part_unregister_skel (vc, argc, argv);
      break;
    case CMD_PART_SETSKELGLOBAL:
      part_register_skel_global (vc, argc, argv);
      break;
    case CMD_PART_DELSKELGLOBAL:
      part_unregister_skel_global (vc, argc, argv);
      break;
    case CMD_PART_SKELLIST:
      part_list_skels (vc, level, argc, argv);
      break;
    case CMD_PART_SETSKELVS:
      part_register_skel_vs (vc, argc, argv);
      break;
    default:
      return 1;
      break;
    }

  return 0;
}
#endif

module part_module =
{
  mod_part_check_cmd,
  mod_part_cmd
};
