/* Control virtual server operation.
   cOPYRIght (c) 1999, 2000, 2001 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>

#include "vsd.h"
#include "part.h"
#include "module.h"
#include "vsd.h"
#include "libvsd.h"

static void send_output (struct connection *vc,
			 const char *server_root, const char *file)
{
  char *temp;
  FILE *stream;
  char line[256];

  if (server_root)
    {
      temp = (char *) malloc (strlen (server_root) + strlen (file) + 3);
      strcpy (temp, server_root);
      strcat (temp, file);
    }
  else
    temp = (char *) file;

  stream = fopen (temp, "r");
  if (stream != NULL)
    {
      while (fgets (line, sizeof (line), stream))
	io_buffer_store (vc, "%s", line);

      fclose (stream);
    }
  unlink (temp);

  if (server_root)
    free (temp);
}

void part_register_skel (struct connection *vc, int argc, char *argv[])
{

  int partition, add_status;
  char *skel = NULL;
  const char *status;

/* PART_SKELADD <skel name> <partition> */

  /* obtain args */
  if ((argv[0] == NULL) || (argv[0] == '\0'))
  {
    io_buffer_store (vc, "-ERR Skel name argument is missing");
  }
  skel = strdup (argv[0]);
  
  if (argc >= 2)
  {
    partition = atoi (argv[1]);
  }
  else 
  {
    io_buffer_store (vc, "-ERR Partition argument is missing");
  }

  /* Is skel name valid? */
  if ((status = part_valid_skelname (skel)))
  {
    io_buffer_store (vc, "-ERR Skel name is invalid: %s", status);
    return;
  }

  /* Create map entry for skel */
  add_status = part_map_add_skel (vc->map, partition, skel);
    if (add_status == 0) /* save map */
    {
       if (vsd_map_save (vc->map))
         io_buffer_store (vc, "-ERR Cannot update vsd.conf: %m");
    }
    else if (add_status == 1)
      io_buffer_store (vc, "-ERR Invalid partition: %d", partition);
    else if (add_status == 2)
      io_buffer_store (vc, "-ERR part_register_skel: partition read problem");
    else if (add_status == 3)
      io_buffer_store (vc, "-ERR Skel already registered");
    else if (add_status == 4)
      io_buffer_store (vc, "-ERR Cannot add %s: memory allocation error", skel);
    else if (add_status == 5)
      io_buffer_store (vc, "-ERR Failed to set new skel value to %s", skel);
   else
     io_buffer_store (vc, "-ERR part_register_skel: unknown error (%d) occurred",
                     add_status);
}

void part_unregister_skel (struct connection *vc, int argc, char *argv[])
{

  int partition, del_status;
  char *skel = NULL;
  const char *status;

/* PART_SKELDEL <skel name> <partition> */

  /* obtain args */
  if ((argv[0] == NULL) || (argv[0] == '\0'))
  {
    io_buffer_store (vc, "-ERR Skel name argument is missing");
  }
  skel = strdup (argv[0]);
 
  if (argc >= 2)
  {
    partition = atoi (argv[1]);
  }
  else
  {
    io_buffer_store (vc, "-ERR Partition argument is missing");
  }

  /* Is skel name valid? */
  if ((status = part_valid_skelname (skel)))
  {
    io_buffer_store (vc, "-ERR Skel name is invalid: %s", status);
    return;
  }

  /* Create map entry for skel */
  del_status = part_map_del_skel (vc->map, partition, skel);
    if (del_status == 0) /* save map */
    {
       if (vsd_map_save (vc->map))
         io_buffer_store (vc, "-ERR Cannot update vsd.conf: %m");
    }
    else if (del_status == 1)
      io_buffer_store (vc, "-ERR Invalid partition: %d", partition);
    else if (del_status == 2)
      io_buffer_store (vc, "-ERR part_register_skel: partition read problem");
    else if (del_status == 3)
      io_buffer_store (vc, "-ERR Skel not registered on partition %d", partition);
    else if (del_status == 4)
      io_buffer_store (vc, "-ERR Cannot unregister %s: memory allocation error", skel);
  else
    io_buffer_store (vc, "-ERR part_unregister_skel: unknown error (%d) occurred",
                     del_status);
}

void part_register_skel_global (struct connection *vc, int argc, char *argv[])
{

/* PART_SETSKELGLOBAL <skel name> */
  const char *status;
  char skel_name[64];

  /* obtain args */
  if (argv[0] == NULL || argv[0] == '\0')
  {
    io_buffer_store (vc, "-ERR Skel name argument is missing");
    return;
  }  
  strcpy (skel_name, argv[0]);

  /* Is the skel name vaild? */
  if ((status = part_valid_skelname (skel_name)))
  {
    io_buffer_store (vc, "-ERR Skel name is invalid: %s", status);
    return;
  }

  /* Is the skel registered on a partition? */
  if (vsd_getskelent (vc->map, skel_name) == NULL)
  {
    io_buffer_store (vc, "-ERR Skel is not registered on any partition");
    return;
  }

  /* Set the skel as global */
  strcpy (vc->map->global.skel, skel_name);

  /* Save map */
  if (vsd_map_save (vc->map))
    io_buffer_store (vc, "-ERR Cannot update vsd.conf: %m");
}

void part_unregister_skel_global (struct connection *vc, int argc, char *argv[])
{

/* PART_DELSKELGLOBAL <skelname> */
  const char *status;
  char skel_name[64];

  /* obtain args */
  if (argv[0] == NULL || argv[0] == '\0')
  {
    io_buffer_store (vc, "-ERR Skel name argument is missing");
    return;
  } 
  strcpy (skel_name, argv[0]);

  /* Is the skel name vaild? */
  if ((status = part_valid_skelname (skel_name)))
  {
    io_buffer_store (vc, "-ERR Skel name is invalid: %s", status);
    return;
  }

  /* Is the global skel aet, and does it match the one passed? */
  if (vc->map->global.skel[0] == '\0')
  {
    io_buffer_store (vc, "-ERR A global skel is not defined");
    return;
  }
  else if (strcmp (vc->map->global.skel, skel_name) != 0)
  {
    io_buffer_store (vc, "-ERR Global skel is not set to %s", skel_name);
    return;
  }
  else 
  {
    /* If so, then delete it */
    vc->map->global.skel[0] = '\0';
    /* Save map */
   if (vsd_map_save (vc->map))
     io_buffer_store (vc, "-ERR Cannot update vsd.conf: %m");
  }
}

void part_list_skels (struct connection *vc, int level, int argc, char *argv[])
{

/* PART_SKELLIST [<partition>] */
/* Returns a list of skels, one line for the global section */
/* and another for each partition */
  
  int part_num, x, y;

  if ((level == ADMIN_VS && argc == 0) || (level == ADMIN_HOST && argc == 1))
  {
    if (argc == 1)
    {
      part_num = atoi (argv[0]);
      /* check that the partition number is valid */
      if (part_num >= vc->map->partitions)
      {
        io_buffer_store (vc, "-ERR Partition number %d is invalid", part_num);
        return;
      }
    }
    else
    {
      /* Get partiton number from virtual server entry */
      /* defined in the connection */
      if (vc->vs == NULL || vc->vs->partition > vc->map->partitions)
      {
        io_buffer_store (vc, "-ERR could not obtain a partition number");
        return;
      }
      else 
      {
        part_num = vc->vs->partition;
      }
  
    }
      /* Print out all skels on that partition */
    io_buffer_store (vc, "partition %d:", part_num);
    for (x = 0; x < vc->map->partition[part_num].skels; x++)
      {
        io_buffer_store (vc, " %s", vc->map->partition[part_num].skelname[x]);
      }
      io_buffer_store (vc, "\n");
    }
  else if (level == ADMIN_HOST)
  {
    /* Print out all information */
    if (vc->map->global.skel[0] != '\0')
    {
      io_buffer_store (vc, "global skel: %s\n", vc->map->global.skel);
    }

    for (x = 0; x < vc->map->partitions; x++)
    {
      io_buffer_store (vc, "partition %d:", x);
      for (y = 0; y < vc->map->partition[x].skels; y++)
      {
        io_buffer_store (vc, " %s", vc->map->partition[x].skelname[y]);
      }
      io_buffer_store (vc, "\n");
    }
  }
  else
  {
    io_buffer_store (vc, "-ERR Permission denied\n");
  }
}

void part_register_skel_vs (struct connection *vc, int argc, char *argv[])
{
/* PART_SETSKELVS <vs name> <skel name> */
  int part_num, x, found_skel = 0;
  const char *status;
  char skel_name[64];
  struct vsd_vs *vs;

  /* obtain args */
  if (argv[0] == NULL || argv[0] == '\0')
  {
    io_buffer_store (vc, "-ERR Virtual server name argument is missing");
    return;
  }
  if ((vs = vsd_getmapent (vc->map, argv[0], NULL)) == NULL)
  {
    io_buffer_store (vc, "-ERR Server %s is not known", argv[0]);
    return;
  }

  if (argv[1] == NULL || argv[1] == '\0')
  {
    io_buffer_store (vc, "-ERR Skel name argument is missing");
    return;
  }
  strcpy (skel_name, argv[1]);

  /* Is the skel name vaild? */
  if ((status = part_valid_skelname (skel_name)))
  {
    io_buffer_store (vc, "-ERR Skel name is invalid: %s", status);
    return;
  }

  /* obtain the partition number of the skel */
  part_num = vs->partition;
  
  /* check that the skel exists on the partition */
  for (x = 0; x < vc->map->partition[part_num].skels; x++)
  {
    if (strcmp (vc->map->partition[part_num].skelname[x], skel_name) == 0)
    {
      /* skel found on partition, so set vs skel */
      strcpy (vs->skel, skel_name);
      found_skel = 1;
    }
  }

  /* check that skel was found, if so, save map, else return error */
  if (found_skel == 1)
  {
    /* save map */
    if (vsd_map_save (vc->map))
      io_buffer_store (vc, "-ERR Cannot update vsd.conf: %m");
      return;
  }
  else
  {
    io_buffer_store (vc, "-ERR Skel not registered on same partition as virtual server");
    return;
  }
  return;
}

/* Take a pointer to a string; check the length and for any invalid characters */
const char *part_valid_skelname (const char *skelname)
{

  if (skelname == NULL)
    return "part_valid_skelname: passed NULL argument";

  if (strlen (skelname) > 64)
    return "skel name too long";

  if (strspn (skelname, "abcdefghijklmnopqrstuvwxyz0123456789-_") != strlen (skelname))
    return "contains unsuitable characters";

  return NULL;
}

/* Add a skel entry to an existing partition.
  Return 0 on success.
  Return 1 if partition does not exist.
  Return 2 if there is a read problem
  Return 3 if skelname is already registered with the partition.
  Return 4 if memory allocation failed. 
  Return 5 if there was a write error */
int part_map_add_skel (struct vsd_vs_map *map,
                      int partition_num, const char *skelname)
{

  struct vsd_disk *part;
  char skel[64];
  int x, y;

  /* Does partition exist? */
  if (partition_num >= map->partitions)
    return 1;

  part = map->partition;

  /* Does a skel with that name already exist on that partition? */
  for (x=0; x < part[partition_num].skels; x++)
  {
    /* Do we need to check that the pointer passed to strcmp is valid? */
    if (part[partition_num].skelname[x] == NULL)
      return 2;

    if (strcmp (skelname, part[partition_num].skelname[x]) == 0)
      return 3;
  }  

  /* Add the skel to the number of skels registered on that partition */
  /* and allocate the memory */
  x = part[partition_num].skels + 1;

  if (x == 1)
    part[partition_num].skelname = malloc (sizeof (skelname));
  else
    part[partition_num].skelname = realloc (part[partition_num].skelname, sizeof (skelname) * x);

  if (part[partition_num].skelname == NULL)
    return 4;

  if ((part[partition_num].skelname[x - 1] = strdup (skelname)) == NULL)
    return 5;

  /* Successfully added, therefore increment the num of skels */
  part[partition_num].skels++;

  return 0;
}

/* Delete a skel entry from an existing partition.
  Return 0 on success.
  Return 1 if partition does not exist.
  Return 2 if there is a read problem
  Return 3 if skelname is not registered with the partition.
  Return 4 if memory allocation failed. */
int part_map_del_skel (struct vsd_vs_map *map,
                      int partition_num, const char *skelname)
{

  struct vsd_disk *part;
  char skel[64]; 
  char **new_skelname;
  int x, y, found_skel = 0;

  /* Does partition exist? */
  if (partition_num >= map->partitions)
    return 1;

  part = &map->partition[partition_num];
  strcpy (skel, skelname);

  /* Quick case: only 1 skel registered */
  if (part->skels == 1)
  {
    if (strcmp (skel, part->skelname[0]) == 0)
    {
      part->skels = 0;
      free (part->skelname);
      part->skelname = NULL;
    }
    else
    {
      return 3;
    }
  }
  else 
  {  
    /* Allocate enough memory to store adjusted array of skel names */
    new_skelname = malloc (sizeof (skel) * (part->skels -1));
    if (new_skelname == NULL)
      return 4;

    /* Does a skel with that name exist on the partition? */
    /* Add the rest of the skels to the new list */
    for (x = y = 0; x < part->skels; x++)
    {
      if (strcmp (part->skelname[x], skel) == 0)
        found_skel = 1;
      else 
        new_skelname[y++] = part->skelname[x];
    }
    if (found_skel != 1)
      return 3;

    free (part->skelname);
    part->skelname = new_skelname;
    part->skels--;  
  }
  
  /* Does a skel with that name exist as a global skel? */
  /* Is the global ske set? */
  if (map->global.skel[0] != '\0')
  {

    /* reset the found skel flag */
    found_skel = 0;

    if (strcmp (skel, map->global.skel) == 0)
    {
      /* If so, check all other partitions for a skel with that name */
      for (x = 0; x < map->partitions; x++)
      {
        for (y = 0; x < map->partition[x].skels
                    && x != partition_num; y++)
        {
          if (strcmp (map->partition[x].skelname[y], skel) == 1)
            found_skel++;
        }
      }
        
      /* If no other partitions, or no other skel entries with same name on other partitions */
      /* then unregister skel global */
      if (found_skel == 0)
        map->global.skel[0] = '\0';
    }
  }
  return 0;
}

